/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.engine.layout.pdf.font;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.osgi.framework.Bundle;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.lowagie.text.FontFactory;

// FIXME: code review : add comments for all methods.
public class FontConfigReader
{

	private static final int DEFAULT_BLOCK_INDEX = -1;
	private static final String TAG_COMPOSITE_FONT = "composite-font"; //$NON-NLS-1$
	private static final String TAG_ALL_FONTS = "all-fonts"; //$NON-NLS-1$
	private static final String TAG_BLOCK = "block"; //$NON-NLS-1$
	private static final String TAG_CHARACTER = "character"; //$NON-NLS-1$
	private static final String TAG_MAPPING = "mapping"; //$NON-NLS-1$
	private static final String TAG_PATH = "path"; //$NON-NLS-1$
	private static final String TAG_ENCODING = "encoding"; //$NON-NLS-1$
	private static final String TAG_FONT_MAPPINGS = "font-mappings"; //$NON-NLS-1$
	private static final String TAG_FONT_ALIASES = "font-aliases"; //$NON-NLS-1$

	private static final String PROP_BLOCK_INDEX = "index"; //$NON-NLS-1$
	private static final String PROP_NAME = "name"; //$NON-NLS-1$
	private static final String PROP_FONT_FAMILY = "font-family"; //$NON-NLS-1$
	private static final String PROP_ENCODING = "encoding"; //$NON-NLS-1$
	private static final String PROP_PATH = "path"; //$NON-NLS-1$
	private static final String DEFAULT_BLOCK = "default";

	/** Prefix of config xml file name */
	private static final String CONFIG_FILE_PREFIX = "/fontsConfig"; //$NON-NLS-1$

	private static final String CONFIG_FILE_SUFFIX = ".xml"; //$NON-NLS-1$

	private static final String CONFIG_FILE_SEPERATOR = "_"; //$NON-NLS-1$

	private FontMappingManager fontMappingManager = new FontMappingManager( );

	/** the logger logging the error, debug, warning messages. */
	protected static Logger logger = Logger.getLogger( FontConfigReader.class
			.getName( ) );

	public void initialize( )
	{
		// read the embeded font path.
		// FIXME: code review : prosses embedded font path using the same method
		// as normal config file paths.
		String embededFontPath = getEmbededFontPath( );
		if ( null != embededFontPath )
		{
			FontFactory.registerDirectory( embededFontPath );
		}
		parseFormatUnrelatedConfigFile( );
	}

	public void parseFormatUnrelatedConfigFile( )
	{
		String defaultConfig = CONFIG_FILE_PREFIX + CONFIG_FILE_SUFFIX;
		parseConfigFile( defaultConfig );

		String platformConfigPrefix = getPlatformConfig( );
		String platformConfig = platformConfigPrefix + CONFIG_FILE_SUFFIX;
		parseConfigFile( platformConfig );

		Locale locale = Locale.getDefault( );
		String languageConfigPrefix = platformConfigPrefix
				+ CONFIG_FILE_SEPERATOR + locale.getLanguage( );
		String languageConfig = languageConfigPrefix + CONFIG_FILE_SUFFIX;
		parseConfigFile( languageConfig );

		// FIXME: code review : consider the case when country is not defined in
		// locale.
		String countryConfig = languageConfigPrefix + CONFIG_FILE_SEPERATOR
				+ locale.getCountry( ) + CONFIG_FILE_SUFFIX;
		parseConfigFile( countryConfig );
	}

	private static String getPlatformConfig( )
	{
		return CONFIG_FILE_PREFIX + CONFIG_FILE_SEPERATOR + getOsName( );
	}

	private static String getOsName( )
	{
		String osName = System.getProperty( "os.name" );
		osName = osName.replaceAll( " ", CONFIG_FILE_SEPERATOR );
		return osName;
	}

	public boolean parseFormatRelatedConfigFile( String format )
	{
		String configFile = CONFIG_FILE_PREFIX + CONFIG_FILE_SEPERATOR + format
				+ CONFIG_FILE_SUFFIX;
		boolean result = parseConfigFile( configFile );
		configFile = CONFIG_FILE_PREFIX + CONFIG_FILE_SEPERATOR + format
				+ CONFIG_FILE_SEPERATOR + getOsName( ) + CONFIG_FILE_SUFFIX;
		return result | parseConfigFile( configFile );
	}

	public boolean parseConfigFile( String configFile )
			throws FactoryConfigurationError
	{
		try
		{
			URL fileURL = getURL( configFile );
			return parseConfigFile( fileURL );
		}
		catch ( Exception se )
		{
			logger.log( Level.WARNING, se.getMessage( ), se );
		}
		return false;
	}

	protected URL getURL( String configFile )
	{
		Bundle bundle = Platform
				.getBundle( "org.eclipse.birt.report.engine.fonts" ); //$NON-NLS-1$
		if ( bundle == null )
			return null;

		URL fileURL = bundle.getEntry( configFile );
		return fileURL;
	}

	public boolean parseConfigFile( URL fileURL ) throws IOException,
			FactoryConfigurationError, ParserConfigurationException,
			SAXException
	{
		if ( null == fileURL )
			return false;

		InputStream cfgFile = fileURL.openStream( );
		try
		{
			InputStreamReader reader = new InputStreamReader(
					new BufferedInputStream( cfgFile ), Charset
							.forName( "UTF-8" ) ); //$NON-NLS-1$
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance( );
			DocumentBuilder db = dbf.newDocumentBuilder( );
			Document doc = db.parse( new InputSource( reader ) );

			handleFontMappings( doc );
			handleCompositeFonts( doc );
			handleFontPaths( doc );
			handleFontEncodings( doc );
		}
		finally
		{
			cfgFile.close( );
		}
		return true;
	}

	public String getEmbededFontPath( )
	{
		Bundle bundle = Platform
				.getBundle( "org.eclipse.birt.report.engine.fonts" ); //$NON-NLS-1$
		Path path = new Path( "/fonts" ); //$NON-NLS-1$

		URL fileURL = FileLocator.find( bundle, path, null );
		if ( null == fileURL )
			return null;
		String fontPath = null;
		try
		{
			// 171369 patch provided by Arne Degenring <public@degenring.de>
			fontPath = FileLocator.toFileURL( fileURL ).getPath( );
			if ( fontPath != null && fontPath.length( ) >= 3
					&& fontPath.charAt( 2 ) == ':' )
			{
				// truncate the first '/';
				return fontPath.substring( 1 );
			}
			else
			{
				return fontPath;
			}
		}
		catch ( IOException ioe )
		{
			logger.log( Level.WARNING, ioe.getMessage( ), ioe );
			return null;
		}
	}

	public FontMappingManager getFontMappingManager( )
	{
		return new FontMappingManager( fontMappingManager );
	}

	public void reset( )
	{
		fontMappingManager.reset( );
	}

	private void handleFontEncodings( Document doc )
	{
		HashMap fontEncoding = new HashMap( );
		NodeList encodings = doc.getDocumentElement( ).getElementsByTagName(
				TAG_ENCODING );
		for ( int i = 0; i < encodings.getLength( ); i++ )
		{
			Node node = encodings.item( i );
			String fontFamily = getProperty( node, PROP_FONT_FAMILY );
			String encoding = getProperty( node, PROP_ENCODING );
			if ( notBlank( encoding ) && notBlank( fontFamily ) )
				fontEncoding.put( fontFamily, encoding );
		}
		fontMappingManager.addFontEncoding( fontEncoding );
	}

	private void handleFontPaths( Document doc )
	{
		NodeList paths = doc.getDocumentElement( ).getElementsByTagName(
				TAG_PATH );
		for ( int i = 0; i < paths.getLength( ); i++ )
		{
			Node node = paths.item( i );
			String path = getProperty( node, PROP_PATH );
			if ( notBlank( path ) )
				fontMappingManager.addFontPath( path );
		}
	}

	private void handleFontMappings( Document doc )
	{
		Element documentElement = doc.getDocumentElement( );
		processFontMappings( documentElement
				.getElementsByTagName( TAG_FONT_MAPPINGS ) );
		processFontMappings( documentElement
				.getElementsByTagName( TAG_FONT_ALIASES ) );
	}

	private void processFontMappings( NodeList fontMappings )
	{
		if ( fontMappings == null || fontMappings.getLength( ) <= 0 )
		{
			return;
		}
		Node fontMapping = fontMappings.item( 0 );
		NodeList childNodes = fontMapping.getChildNodes( );
		HashMap fontMappingResult = new HashMap( );
		for ( int i = 0; i < childNodes.getLength( ); i++ )
		{
			Node node = childNodes.item( i );
			if ( node.getNodeType( ) == Node.ELEMENT_NODE
					&& TAG_MAPPING.equals( node.getNodeName( ) ) )
			{
				String name = getProperty( node, PROP_NAME );
				String fontFamily = getProperty( node, PROP_FONT_FAMILY );
				if ( notBlank( name ) && notBlank( fontFamily ) )
					fontMappingResult.put( name, fontFamily );
			}
		}

		fontMappingManager.addFontMapping( fontMappingResult );
	}

	private void handleCompositeFonts( Document doc )
	{
		handleAllFontsNode( doc );
		handleCompositeFontsNode( doc );
	}

	private void handleCompositeFontsNode( Document doc )
	{
		NodeList compositeFonts = doc.getDocumentElement( )
				.getElementsByTagName( TAG_COMPOSITE_FONT );
		for ( int i = 0; i < compositeFonts.getLength( ); i++ )
		{
			Node node = compositeFonts.item( i );
			CompositeFont compositeFont = parseComsiteFont( node );
			fontMappingManager.addCompositeFonts( compositeFont.getName( ),
					compositeFont );
		}
	}

	private void handleAllFontsNode( Document doc )
	{
		NodeList allFonts = doc.getDocumentElement( ).getElementsByTagName(
				TAG_ALL_FONTS );
		if ( allFonts.getLength( ) != 0 )
		{
			Node node = allFonts.item( 0 );
			NodeList blocks = node.getChildNodes( );
			for ( int i = 0; i < blocks.getLength( ); i++ )
			{
				Node blockNode = blocks.item( i );
				if ( !TAG_BLOCK.equalsIgnoreCase( blockNode.getNodeName( ) ) )
				{
					continue;
				}
				String blockIndex = getProperty( blockNode, PROP_BLOCK_INDEX );
				if ( notBlank( blockIndex ) )
				{
					int index = Integer.parseInt( blockIndex );
					processAllFontsMapping( blockNode, index );
					continue;
				}
				String blockName = getProperty( blockNode, PROP_NAME );
				if ( DEFAULT_BLOCK.equalsIgnoreCase( blockName ) )
				{
					processAllFontsMapping( blockNode, DEFAULT_BLOCK_INDEX );
				}
			}
		}
	}

	private void processAllFontsMapping( Node blockNode, int blockIndex )
	{
		boolean isDefault = ( blockIndex == DEFAULT_BLOCK_INDEX );
		NodeList mappings = blockNode.getChildNodes( );
		for ( int j = 0; j < mappings.getLength( ); j++ )
		{

			Node mapping = mappings.item( j );
			if ( mapping.getNodeType( ) != Node.ELEMENT_NODE )
				continue;

			String name = getProperty( mapping, PROP_NAME );
			String fontFamily = getProperty( mapping, PROP_FONT_FAMILY );
			if ( notBlank( fontFamily ) )
			{
				if ( isDefault )
				{
					fontMappingManager.setDefaultFont( name, fontFamily );
				}
				else
				{
					fontMappingManager.addBlockToCompositeFont( name,
							blockIndex, fontFamily );
				}
			}
		}
	}

	private CompositeFont parseComsiteFont( Node node )
	{
		NodeList blocks = node.getChildNodes( );
		String fontName = getProperty( node, PROP_NAME );
		String defaultFont = getProperty( node, PROP_FONT_FAMILY );
		CompositeFont compositeFont = new CompositeFont( fontName, defaultFont );
		for ( int i = 0; i < blocks.getLength( ); i++ )
		{
			Node blockNode = blocks.item( i );
			String nodeName = blockNode.getNodeName( );
			if ( TAG_BLOCK.equalsIgnoreCase( nodeName ) )
			{
				String blockIndex = getProperty( blockNode, PROP_BLOCK_INDEX );
				if ( notBlank( blockIndex ) )
				{
					int index = Integer.parseInt( blockIndex );
					String font = getProperty( blockNode, PROP_FONT_FAMILY );
					compositeFont.setBlockFont( index, font );
					continue;
				}

				// Compatible for older version which uses block named "default"
				// to define default font for the composite font. Current
				// implementation use attribute "default" defined on
				// "composite-font" to define default font for the composite
				// font.
				String blockName = getProperty( blockNode, PROP_NAME );
				if ( DEFAULT_BLOCK.equalsIgnoreCase( blockName ) )
				{
					String font = getProperty( blockNode, PROP_FONT_FAMILY );
					if ( font != null
							&& compositeFont.getDefaultFont( ) == null )
					{
						compositeFont.setDefaultFont( font );
					}
				}
			}
			else if ( TAG_CHARACTER.equalsIgnoreCase( nodeName ) )
			{
				String character = getProperty( blockNode, "value" );
				try
				{
					char c = getChar( character );
					String font = getProperty( blockNode, PROP_FONT_FAMILY );
					if ( font != null )
					{
						compositeFont.setCharacterFont( c, font );
					}
				}
				catch ( Exception e )
				{
					logger.log( Level.WARNING, e.getMessage( ), e );
				}
			}
		}
		return compositeFont;
	}

	/**
	 * Parses a character from a string represents the character. The string can
	 * be of two form: character and UNICODE. For instance, character 'a' can be
	 * represented by "a" as character form, or "\u0061" as UNICODE form.
	 * 
	 * @param character
	 *            the string represent the character.
	 * @throws Exception
	 *             when parameter character is not a valid character
	 *             representation.
	 */
	private char getChar( String character ) throws Exception
	{
		if ( character.matches( "." ) )
		{
			return character.charAt( 0 );
		}
		else if ( character.matches( "\\\\u\\p{XDigit}{4}" ) )
		{
			String unicode = character.substring( 2 );
			return (char) Integer.parseInt( unicode, 16 );
		}
		throw new Exception( "Unknow character value: " + character );
	}

	private boolean notBlank( String propertyName )
	{
		return ( null != propertyName && propertyName.length( ) != 0 );
	}

	private String getProperty( Node node, String propertyName )
	{
		if ( null == node )
			return null;
		NamedNodeMap atts = node.getAttributes( );
		if ( null != atts )
		{
			Node property = atts.getNamedItem( propertyName );
			if ( null != property )
				return property.getNodeValue( );
		}
		return null;
	}
}
